#!/usr/bin/env python
# coding=utf8

import copy
import json
from pprint import pprint
import symmetry

downmoves = set('aghij4lmno 5')
def coolness(word):
    totaldown=0
    for i in word:
        if i in downmoves:
            totaldown+=1
    return float(len(word))/totaldown

WORDS_OF_POWER = [
    'ei!',
    "r'lyeh",
    'ia! ia!',
    'yuggoth',
    'planet 10',
    'tsathoggua',
    'yogsothoth',
    'john bigboote',
    'cthulhu fhtagn!',
    "Ph'nglui mglw'nafh Cthulhu R'lyeh wgah'nagl fhtagn."
]

W = 'p'
E = 'b'
SW = 'a'
SE = 'l'
CW = 'd'
CCW = 'k'
NOP = '\t'

CMD_TO_MUTATIONS = {
    W: "p'!.03",
    E: 'bcefy2',
    SW: 'aghij4',
    SE: 'lmno 5',
    CW: 'dqrvz1',
    CCW: 'kstuwx',
    NOP: '\t\n\r'
}

MUTATION_TO_CMD = {m: c
                   for c, ms in CMD_TO_MUTATIONS.items()
                   for m in ms + ms.upper()}


# See http://www.redblobgames.com/grids/hexagons/
def offset_to_cube(c, r):
    x = c - (r - (r & 1)) / 2
    z = r
    y = -x - z
    return (x, y, z)


def cube_to_offset(x, y, z):
    c = x + (z - (z & 1)) / 2
    r = z
    return c, r


def rotate((pc, pr), (c, r), rot):
    px, py, pz = offset_to_cube(pc, pr)
    x, y, z = offset_to_cube(c, r)
    dx, dy, dz = (x - px, y - py, z - pz)
    for _ in xrange(rot):
        [dx, dy, dz] = [-dz, -dx, -dy]

    return cube_to_offset(px + dx, py + dy, pz + dz)


def move_point(step, (x, y), pivot_xy):
    if step == W:
        return (x - 1, y)
    elif step == E:
        return (x + 1, y)
    elif step == SW:
        return (x - 1 + (y & 1), y + 1)
    elif step == SE:
        return (x + (y & 1), y + 1)
    elif step == CW:
        return rotate(pivot_xy, (x, y), 1)
    elif step == CCW:
        return rotate(pivot_xy, (x, y), 5)
    else:
        raise Exception("unknown cmd %s" % repr(step))


def reverse_step(step, x, y, rot, sym):
    if step == W:
        return (x + 1, y, rot)
    elif step == E:
        return (x - 1, y, rot)
    elif step == SW:
        return (x + (y & 1), y - 1, rot)
    elif step == SE:
        return (x - 1 + (y & 1), y - 1, rot)
    elif step == CW:
        return (x, y, (rot - 1) % sym)
    elif step == CCW:
        return (x, y, (rot + 1) % sym)
    else:
        raise Exception("unknown cmd %s" % repr(step))


def forward_step(step, x, y, rot, sym):
    if step == W:
        return (x - 1, y, rot)
    elif step == E:
        return (x + 1, y, rot)
    elif step == SW:
        return (x - 1 + (y & 1), y + 1, rot)
    elif step == SE:
        return (x + (y & 1), y + 1, rot)
    elif step == CW:
        return (x, y, (rot + 1) % sym)
    elif step == CCW:
        return (x, y, (rot - 1) % sym)
    elif step == NOP:
        return (x, y, rot)
    else:
        raise Exception("unknown cmd %s" % repr(step))


class Unit(object):
    def __init__(self, pivot_xy, members, w, h, _force_sym=0):
        self.w = w
        self.h = h
        self.bmaps = [0L] * 2 * 6
        self.offsets = []  # 12 elems
        self.bound_rect = []  # 12 elems, [(x0, y0, x1, y1), ...]
        self.pivot = (pivot_xy[0], pivot_xy[1])
        self.size = len(members)
        self.mass_center_x = [0] * 6
        self.mass_center_y = [0] * 6
        self.members = members

        rotated_set = set()
        for rot in xrange(_force_sym or 6):
            rotated = [rotate(self.pivot, xy, rot) for xy in members]
            rotated_set.add(tuple(sorted(rotated)))

            flen = float(len(rotated))
            self.mass_center_x[rot] = sum(x for x, _ in rotated) / flen
            self.mass_center_y[rot] = sum(y for _, y in rotated) / flen

            off = min(x + y * w for x, y in rotated)
            off_odd = min(x + (y & 1) + y * w for x, y in rotated)
            self.offsets.append(off)
            self.offsets.append(off_odd)
            for (x, y) in rotated:
                self.bmaps[0 + 2 * rot] |= 1 << (x + y * w - off)
                self.bmaps[1 + 2 * rot] |= 1 << (x + (y & 1) + y * w - off_odd)

            y_min = min(y for _, y in rotated)
            y_max = max(y for _, y in rotated)
            x_min = min(x for x, _ in rotated)
            x_max = max(x for x, _ in rotated)
            x_min_odd = min(x + (y & 1) for x, y in rotated)
            x_max_odd = max(x + (y & 1) for x, y in rotated)
            self.bound_rect.append((x_min, y_min, x_max, y_max))
            self.bound_rect.append((x_min_odd, y_min, x_max_odd, y_max))

        self.padding = (
            -min(x1 for x1, _, _, _ in self.bound_rect),
            -min(y1 for _, y1, _, _ in self.bound_rect),
            +max(x2 for _, _, x2, _ in self.bound_rect),
            +max(y2 for _, _, _, y2 in self.bound_rect),
        )
        self.span=self.padding[2]-self.padding[0]+self.padding[3]-self.padding[1]

        self.symmetry = len(rotated_set)
        self._set_spawn_coords()

    def trans_is_valid(self, dx, dy, rot):
        odd = dy & 1
        x_min, y_min, x_max, y_max = self.bound_rect[odd + 2 * rot]
        return ((-x_min <= dx < self.w - x_max) and
                (-y_min <= dy < self.h - y_max))

    def trans_bmap(self, dx, dy, rot):
        # FIXME: return None if out-of-bound?
        i = (dy & 1) + 2 * rot
        return self.bmaps[i] << (dx + dy*self.w + self.offsets[i])

    def trans_pivot(self, dx, dy):
        if self.pivot[1] & 1:
            dx += dy & 1
        return (self.pivot[0] + dx, self.pivot[1] + dy)

    def get_bound_rect(self, dy, rot):
        odd = dy & 1
        return self.bound_rect[odd + 2 * rot]

    def get_x_mass_center(self, rot):
        return self.mass_center_x[rot]

    def get_y_mass_center(self, rot):
        return self.mass_center_y[rot]

    def _set_spawn_coords(self):
        """Returns (x, y, rot). If there is no space to spawn,
           returns None"""
        rot = 0
        xmin, ymin, xmax, ymax = self.get_bound_rect(0, rot)
        y_offset = -ymin
        if y_offset:
            # print 'y_offset is not zero!!!', y_offset # FIXME?
            xmin, ymin, xmax, ymax = self.get_bound_rect(y_offset, rot)

        # FIXME: proper horizontal alignment
        sidespace = self.w - (xmax - xmin + 1)
        left_offset = sidespace / 2
        x_offset = left_offset - xmin
        self.spawn_coords = x_offset, y_offset, rot


class Board(object):
    def __init__(self, w, h, filled=[]):
        self.bmap = 0L
        self.w = w
        self.h = h

        for x, y in filled:
            self.bmap |= 1 << (x + y * self.w)

    def fill_cell(self, x, y):
        self.bmap |= 1 << (x + y * self.w)

    def show(self, pivot=None):
        lines = ['--' * (self.w + 1)]
        for y in xrange(self.h):
            line = ' >' if y & 1 else '<'
            for x in xrange(self.w):
                i = x + y * self.w
                filled = self.bmap & (1 << i)
                if (x, y) != pivot:
                    line += '◄►' if filled else '· '
                else:
                    line += '◘◘' if filled else '••'
            lines.append(line + ('>' if y & 1 else '<'))
        lines.append('--' * (self.w + 1))
        return '\n'.join(lines)

    def lock(self, unit_bmap):
        self.bmap |= unit_bmap

    def eliminate(self):
        n = 0
        full = (1 << self.w) - 1
        for y in xrange(self.h):
            line = (self.bmap >> self.w * y) & full
            if line == full:
                n += 1
                immov = self.bmap & ((2**(self.w*(self.h-y-1))-1) << self.w*(y + 1))
                mov = (self.bmap & (2**(y*self.w)-1)) << self.w

                self.bmap = immov | mov
        return n

    def unit_fits(self, unit, x, y, rot):
        if not unit.trans_is_valid(x, y, rot):  # inside wall
            return False

        u_bmap = unit.trans_bmap(x, y, rot)
        return not bool(self.bmap & u_bmap)

    def show_with_unit(self, unit, x, y, rot):
        bmap0 = self.bmap
        self.lock(unit.trans_bmap(x, y, rot))
        s = self.show(pivot=unit.trans_pivot(x, y))
        self.bmap = bmap0
        return s

class RandGen(object):
    def __init__(self, seed):
        self.cache = [seed]

    def initial_seed(self):
        return self.cache[0]

    def _prepare_cache(self, index):
        n = index + 1 - len(self.cache)
        if n > 0:
            s = self.cache[-1]
            for _ in xrange(n):
                s = (1103515245 * s + 12345) & 0xffffffff
                self.cache.append(s)

    def __getitem__(self, i):
        self._prepare_cache(i)
        return (self.cache[i] >> 16) & 0x7fff


class WordsOfPower(object):
    def __init__(self, words):
        self.words = list(words)

        self.sym = {1: [], 2: [], 3: [], 6: []}
        for w in self.words:
            sym = symmetry.reqsym(w)
            cool = coolness(w)
            for j in self.sym.keys():
                if j >= sym:
                    self.sym[j].append((cool, w))

        for j in self.sym.keys():
            # prefer coolest, than short
            self.sym[j].sort(key=lambda (c, w): (-c, len(w)))

        self.spam = {}
        for s, c_ws in self.sym.items():
            w2 = [w for _, w in c_ws[:2]]
            w_cool = [w for c, w in c_ws[2:] if c > 2.5]
            # w2 = []
            # w_cool = [w for c, w in c_ws if c > 2.45]
            self.spam[s] = w2 + w_cool

    def filter(self, words, sym):
        possible = set(w for _, w in self.sym[sym])
        return [w for w in words if w in possible]


class Game(object):
    FIXED = 'fixed'
    MOVED = 'moved'
    FINISHED = 'finished'

    def __init__(self, board, units, source_length, source_seed):
        # static vars
        self.units = units
        self.source_length = source_length
        self.rng = RandGen(source_seed)

        # current state
        self.board = board
        self.unit_num = None  # number of unit/round
        self.unit_idx = None
        self.unit_pos = None
        self._spawn_new_unit()


    def _spawn_new_unit(self):
        self.unit_num = 0 if self.unit_num is None else self.unit_num + 1
        if self.unit_num >= self.source_length:
            self.unit_idx = None
            return

        self.unit_idx = self.rng[self.unit_num] % len(self.units)
        self.unit_pos = self.units[self.unit_idx].spawn_coords
        x, y, rot = self.unit_pos
        if not self.board.unit_fits(self.units[self.unit_idx], x, y, rot):
            self.unit_idx = None

    def show_with_current_unit(self):
        if self.unit_idx is None:
            return self.board.show()
        else:
            x, y, rot = self.unit_pos
            return self.show_with_unit(self.units[self.unit_idx], x, y, rot)

    def show_with_unit(self, unit, x, y, rot):
        bmap0 = self.board.bmap
        self.board.lock(unit.trans_bmap(x, y, rot))
        s = self.board.show(pivot=unit.trans_pivot(x, y))
        self.board.bmap = bmap0
        return s

    def step(self, mut_cmd):
        """cmd is one of E, W, SE, ....
           Return value is Game.SPAWNED, Game.MOVED, Game.FINISHED"""
        cmd = MUTATION_TO_CMD[mut_cmd]
        if self.unit_idx is None:
            return (self.FINISHED, None)

        unit = self.units[self.unit_idx]
        x, y, rot = self.unit_pos
        x1, y1, rot1 = forward_step(cmd, x, y, rot, unit.symmetry)
        if self.board.unit_fits(unit, x1, y1, rot1):
            self.unit_pos = x1, y1, rot1
            return (self.MOVED, None)
        else:
            self.board.lock(unit.trans_bmap(x, y, rot))
            lines = self.board.eliminate()
            self._spawn_new_unit()
            status = self.FINISHED if self.unit_idx is None else self.FIXED
            return (status, (unit, lines))

    def exec_and_score(self, moves):
        ls_old = 0
        move_scores = 0
        for m_idx, m in enumerate(moves):
            mm = MUTATION_TO_CMD[m]
            status, args = self.step(MUTATION_TO_CMD[m])
            ls = 0
            if status in (self.FIXED, self.FINISHED):
                unit, ls = args
                points = unit.size + 100 * (1 + ls) * ls / 2
                line_bonus = (ls_old - 1) * points / 10 if ls_old else 0
                move_score = points + line_bonus
                move_scores += move_score
            if status == self.FINISHED:
                if m_idx + 1 != len(moves):
                    # TODO: raise?
                    return 0
            ls_old = ls

        power_scores = 0
        for w in WORDS_OF_POWER:
            n = moves.count(w)
            pscore = 2 * len(w) * n
            pbonus = 300 if n > 0 else 0
            power_scores += pscore + pbonus

        # TODO: words of power
        return move_scores, power_scores


def power_score_and_used(moves, words, already_used):
    power_scores = 0
    used_now = set()
    for w in words:
        n = moves.count(w)
        pscore = 2 * len(w) * n
        pbonus = 300 if (n > 0 and w not in already_used)  else 0
        power_scores += pscore + pbonus
        if n and w not in already_used:
            used_now.add(w)
    return power_scores, used_now


def is_end_of_game(board, unit, unit_num, source_length):
    if unit_num >= source_length:
        return True
    x, y, rot = unit.spawn_coords
    return not board.unit_fits(unit, x, y, rot)


def power_rewrite(moves):
    for word in WORDS_OF_POWER:
        canonical = ''.join(MUTATION_TO_CMD[c] for c in word)
        moves = moves.replace(canonical, word)
        # print word, canonical
    return moves


class Games(object):
    def __init__(self, fname):
        to_tuples = lambda coord_list: [(c['x'], c['y']) for c in coord_list]

        with file(fname) as f:
            js = json.load(f)

        # pprint(js)
        w = js['width']
        h = js['height']

        self.board = Board(w, h, to_tuples(js['filled']))
        self.units = [Unit((unit['pivot']['x'], unit['pivot']['y']),
                      to_tuples(unit['members']), w, h)
                 for unit in js['units']]
        self.source_length = js['sourceLength']
        self.seeds = js['sourceSeeds']
        self.task_id = js['id']

    def games_n(self):
        return len(self.seeds)

    def new_game(self, i):
        _board = copy.copy(self.board)
        return Game(_board, self.units, self.source_length, self.seeds[i])


if __name__ == '__main__':
    import sys
    from util import getch

    for x in range(5):
        for y in range(5):
            for rot in range(3):
                for s in [W, E, SW, SE, CW, CCW]:
                    x1, y1, rot1 = forward_step(s, x, y, rot, 3)
                    x2, y2, rot2 = reverse_step(s, x1, y1, rot1, 3)
                    if (x, y, rot) != (x2, y2, rot2):
                        raise Exception("%s != %s" % ((x, y, rot), (x2, y2, rot2)))

    for w in WORDS_OF_POWER:
        x, y, rot = 0, 0, 0
        for c in w:
            x, y, rot = forward_step(MUTATION_TO_CMD[c], x, y, rot, 6)
        print '%2d %2d %2d len=%d %s' % (x, y, rot, len(w), w)

    wop = WordsOfPower(WORDS_OF_POWER)
    for sym, words in wop.sym.items():
        print sym, wop.spam[sym]
        print sym, words
        print

    if len(sys.argv) > 1:
        gs = Games(sys.argv[1])
        g = gs.new_game(0)

        rot = 0
        dx, dy = 0, 0

        bmap0 = g.board.bmap
        for i, u in enumerate(g.units):
            print u.padding
            while 1:
                print i, 'sym=%d' % u.symmetry, (dx, dy, rot)
                print g.show_with_unit(u, dx, dy, rot)

                cmd = getch()
                dx_new, dy_new, rot_new = dx, dy, rot
                if cmd == 'k':
                    dy_new -= 1
                elif cmd == 'j':
                    dy_new += 1
                elif cmd == 'h':
                    dx_new -= 1
                elif cmd == 'l':
                    dx_new += 1
                elif cmd == 'f':
                    rot_new = (rot + 1) % u.symmetry
                elif cmd == 'd':
                    break
                elif cmd == '\x03':
                    sys.exit(-1)
                else:
                    print repr(cmd)

                if u.trans_is_valid(dx_new, dy_new, rot_new):
                    dx, dy, rot = dx_new, dy_new, rot_new
                else:
                    print 'Invalid transform'
