#!/usr/bin/env pypy
from copy import copy

from game import Board, Unit, WORDS_OF_POWER, MUTATION_TO_CMD, forward_step
from power import PowerPattern
from symmetry import reqsym
from wave import ReverseWave


class Patterns(object):
    def __init__(self, units, words):
        self.units = units
        self.words = words
        pp = []
        for u_idx, u in enumerate(units):
            pp.append({word: PowerPattern(u, word) for word in words})
        self.pp = pp

    def get_all_for_unit_idx(self, unit_idx):
        return self.pp[unit_idx]


class TraceCache(object):
    def __init__(self, pos0, unit_symmetry):
        self.pos0 = pos0
        self.sym = unit_symmetry
        self.cached_path = []
        self.cached_moves = ''

    def pos_at_fixme(self, move_idx, moves):
        # FIXME: implement me
        """ 0 <= move_idx <= len(moves) """
        good_i = 0
        for a, b in zip(self.cached_moves, moves):
            if a == b:
                good_i += 1
            else:
                break

    def pos_at(self, move_idx, moves):
        """ 0 <= move_idx <= len(moves) """
        pos = self.pos0
        for i in range(move_idx):
            c = MUTATION_TO_CMD[moves[i]]
            pos = forward_step(c, pos[0], pos[1], pos[2], self.sym)
        return pos


class Mutator(object):
    def __init__(self, patterns, board, unit, moves):
        self.patterns = patterns
        self.board = board
        self.unit = unit
        self.moves0 = moves # need them?
        self.moves = moves
        self.trace_cache = TraceCache(unit.spawn_coords, unit.symmetry)

        end_pos = self.trace_cache.pos_at(len(moves), moves)
        self.rev_wave = ReverseWave(board, unit, *end_pos)

    def try_mutate_at(self, i, word):
        """Try to mutate `moves`, inserting given word at index i.
           Returns: None on fail
                    on success - index of next letter after patch"""
        if i >= len(self.moves):
            return None

        # check if complex movement fits on map
        start_pos = self.trace_cache.pos_at(i, self.moves)
        pattern = self.patterns[word]
        if not self.board.unit_fits(pattern, *start_pos):
            return None

        # add power word and check if destination is still reachable
        moves = self.moves[:i] + word
        last_subst_idx = len(moves)
        end_pos = self.trace_cache.pos_at(len(moves), moves)
        if not self.rev_wave.point_is_reachable(*end_pos):
            return None


        # add path to reach endpoint and check if sequence is valid
        # TODO try different path variations
        rest = self.rev_wave.path_from(*end_pos)
        moves += rest
        sym_needed = reqsym(moves)
        # print 'reqsym', sym_needed
        if self.unit.symmetry < sym_needed:
            return None

        # print moves
        self.moves = moves

        board = copy(self.board)
        # for y in range(board.h):
        #     for x in range(board.w):
        #         if self.rev_wave.point_is_reachable(x, y, 0):
        #             board.fill_cell(x, y)
        # print board.show()
        return last_subst_idx

    def put_any_word_at(self, i, words):
        for w in words:
            pos = self.try_mutate_at(i, w)
            if pos is not None:
                return pos, w
        return None, None

    def spam(self, i, words):
        last_subst_idx = None
        while 1:
            pos, _ = self.put_any_word_at(i, words)
            if pos is None:
                i += 1
            else:
                i = pos
                last_subst_idx = pos
            if i >= len(self.moves):
                break
        return last_subst_idx

    def put_any_word(self, words):
        for i in range(len(self.moves)):
            pos, w = self.put_any_word_at(i, words)
            if pos is not None:
                return pos, w
        return None, None

    def put_one_then_spam(self, words, spamwords):
        pos, w = self.put_any_word(words)
        pos = self.spam(pos or 0, spamwords)
        return pos, w

    def sanity_check(self):
        end_pos0 = self.trace_cache.pos_at(len(self.moves0), self.moves0)
        end_pos = self.trace_cache.pos_at(len(self.moves), self.moves)
        if end_pos0 != end_pos:
            raise Exception("distinations differ")



def main():
    import game
    games = game.Games('qualifiers/problem_01.json')
    board = Board(15, 30)
    # board = games.board
    units = [Unit((0,0), [(0,0)], board.w, board.h)]
    patterns = Patterns(units, WORDS_OF_POWER)
    unit_idx = 0

    longwords = [w for w in WORDS_OF_POWER if w != 'ei!']
    # moves_orig = "ppppaaaaaalallb"
    moves_orig = "aaaaaalllllllllllllllllallb"
    mt = Mutator(patterns.get_all_for_unit_idx(unit_idx),
                 board, units[unit_idx], moves_orig)

    #mt.try_mutate_at(0, 'yuggoth')
    # print mt.try_mutate_at(0, 'ei!')
    # print mt.spam(0, ['ei!'])
    print mt.put_one_then_spam(longwords, ['ei!', "r'lyeh"])
    print mt.moves

    for i in range(len(mt.moves) + 1):
        pos = mt.trace_cache.pos_at(i, mt.moves)
        print board.show_with_unit(units[unit_idx], *pos)
        raw_input()

    mt.sanity_check()

if __name__ == '__main__':
    main()
