#!/usr/bin/python
import array
from collections import deque

from game import E, W, SE, SW, CW, CCW, reverse_step, forward_step


def possible_steps(x, y, rot, sym):
    yield (W, x - 1, y, rot)
    yield (E, x + 1, y, rot)

    yield (SW, x + (y & 1) - 1, y + 1, rot)
    yield (SE, x + (y & 1), y + 1, rot)

    if sym > 1:
        yield (CW, x, y, (rot + 1) % sym)
    if sym > 2:
        yield (CCW, x, y, (rot - 1) % sym)


def possible_steps_ref(x0, y0, rot0, sym):
    cmds = [W, E, SW, SE]
    if sym > 1:
        cmds.append(CW)
    if sym > 2:
        cmds.append(CCW)

    for cmd in  cmds:
        x, y, rot = forward_step(cmd, x0, y0, rot0, sym)
        yield (cmd, x, y, rot)


class SimpleWave(object):
    def __init__(self, board, unit, x_orig, y_orig, rot_orig):
        px1, py1, px2, py2 = unit.padding
        self.w = board.w + px1 + px2
        self.h = board.h + py1 + py2
        self.offset = px1 + py1 * self.w
        self.sym = unit.symmetry
        self._trace_all(board, unit, x_orig, y_orig, rot_orig)

    def _trace_all(self, board, unit, x_orig, y_orig, rot_orig):
        positions = set()

        sym = unit.symmetry
        h, w = self.h, self.w
        off = self.offset

        a = array.array('l', [-1] * (w * h * sym))
        step_map = array.array('c', ['\0'] * (w * h * sym))

        a[off + x_orig + y_orig * w + rot_orig * w * h] = 0
        q = deque([(x_orig, y_orig, rot_orig)])
        while q:
            x0, y0, rot0 = q.popleft()
            dist0 = a[off + x0 + y0 * w + rot0 * w * h]
            # replacing generator with list + forward step leads to 60%
            # slow down, so not generic forward/reverse verstion
            for step, x, y, rot in possible_steps(x0, y0, rot0, sym):
                if board.unit_fits(unit, x, y, rot):
                    # print board.show()
                    # print board.show_with_unit(unit, x, y, rot)
                    # print x, y, rot
                    i = off + x + y * w + rot * w * h
                    # print w, h, i
                    if a[i] < 0 or a[i] > dist0 + 1:
                        a[i] = dist0 + 1
                        step_map[i] = step
                        q.append((x, y, rot))
                else:
                    positions.add((x0, y0, rot0, step))
        self.positions = positions
        self.step_map = step_map

    def get_positions(self):
        return self.positions

    def get_unique_positions(self):
        d = {(x, y, rot): step for x, y, rot, step in self.positions}
        return [(x, y, rot, step) for (x, y, rot), step in d.items()]

    def path_to(self, x, y, rot):
        path = []
        off = self.offset
        while 1:
            i = off + x + y * self.w + rot * self.w * self.h
            step = self.step_map[i]
            if step == '\0':
                path.reverse()
                return ''.join(path)
            else:
                path.append(step)
                x, y, rot = reverse_step(step, x, y, rot, self.sym)


class ReverseWave(SimpleWave):
    # code duplication, see SimpleWave comment for the reason
    def _trace_all(self, board, unit, x_orig, y_orig, rot_orig):
        positions = set()

        sym = unit.symmetry
        h, w = self.h, self.w
        off = self.offset

        a = array.array('l', [-1] * (w * h * sym))
        step_map = array.array('c', ['\0'] * (w * h * sym))

        a[off + x_orig + y_orig * w + rot_orig * w * h] = 0
        q = deque([(x_orig, y_orig, rot_orig)])
        while q:
            x0, y0, rot0 = q.popleft()
            dist0 = a[off + x0 + y0 * w + rot0 * w * h]
            for step in [E, W, SE, SW, CW, CCW]:
                x, y, rot = reverse_step(step, x0, y0, rot0, sym)
                if board.unit_fits(unit, x, y, rot):
                    # print board.show()
                    # print board.show_with_unit(unit, x, y, rot)
                    # print x, y, rot
                    i = off + x + y * w + rot * w * h
                    # print w, h, i
                    if a[i] < 0 or a[i] > dist0 + 1:
                        a[i] = dist0 + 1
                        step_map[i] = step
                        q.append((x, y, rot))
                else:
                    positions.add((x0, y0, rot0, step))  # useless, remove?
        self.arr = a
        self.positions = positions
        self.step_map = step_map

    def path_from(self, x, y, rot):
        path = []
        off = self.offset
        while 1:
            i = off + x + y * self.w + rot * self.w * self.h
            step = self.step_map[i]
            if step == '\0':
                return ''.join(path)
            else:
                path.append(step)
                x, y, rot = forward_step(step, x, y, rot, self.sym)

    def point_is_reachable(self, x, y, rot):
        i = self.offset + x + y * self.w + rot * self.w * self.h
        return self.arr[i] >= 0


if __name__ == '__main__':
    import sys
    import time
    from game import Games

    g = Games(sys.argv[1]).new_game(0)

    # WClass = ReverseWave
    WClass = SimpleWave
    dt_sum = 0

    # pypy warmup
    t0 = time.time()
    for i, unit in enumerate(g.units):
        w = WClass(g.board, unit, 0, 0, 0)
    dt = time.time() - t0
    print 'Warmup time: %.6fs' % dt


    for i, unit in enumerate(g.units):
        for dy in range(4):
            for r in range(unit.symmetry):
                st1 = list(possible_steps(2, dy, r, unit.symmetry))
                st2 = list(possible_steps_ref(2, dy, r, unit.symmetry))
                if st1 != st2:
                    raise Exception("\n%s != \n%s" % (st1, st2))

        t0 = time.time()
        w = WClass(g.board, unit, 0, 0, 0)
        dt = time.time() - t0
        dt_sum += dt
        print 'unit=%d, %d pos, %.6fs' % (i, len(w.get_positions()), dt)
    print '\nTotal time: %.6fs\n' % dt_sum

    bmap0 = g.board.bmap
    for i, unit in enumerate(g.units):
        w = WClass(g.board, unit, *unit.spawn_coords)
        positions = list(w.get_unique_positions())
        print '-' * 50, 'unit %d, %d pos, %.6fs' % (i, len(positions), dt)
        positions.sort(key=lambda (x, y, r, s): (r, y, x))
        for x, y, rot, step in positions:
            print x, y, rot, step
            print w.path_to(x, y, rot)
            g.board.lock(unit.trans_bmap(x, y, rot))
            print g.board.show(pivot=unit.trans_pivot(x, y))
            g.board.bmap = bmap0
            if raw_input().strip():
                break
